@@ -25,7 +25,8 @@ from gridfs.errors import NoFile
 from gridfs.grid_file import (GridIn,
                               GridOut,
                               GridOutCursor,
-                              DEFAULT_CHUNK_SIZE)
+                              DEFAULT_CHUNK_SIZE,
+                              _clear_entity_type_registry)
 from pymongo import (ASCENDING,
                      DESCENDING)
 from pymongo.common import UNAUTHORIZED_CODES, validate_string
@@ -61,6 +62,8 @@ class GridFS(object):
         if not isinstance(database, Database):
             raise TypeError("database must be an instance of Database")
 
+        database = _clear_entity_type_registry(database)
+
         if not database.write_concern.acknowledged:
             raise ConfigurationError('database must use '
                                      'acknowledged write_concern')
@@ -443,6 +446,8 @@ class GridFSBucket(object):
         if not isinstance(db, Database):
             raise TypeError("database must be an instance of Database")
 
+        db = _clear_entity_type_registry(db)
+
         wtc = write_concern if write_concern is not None else db.write_concern
         if not wtc.acknowledged:
             raise ConfigurationError('write concern must be acknowledged')
@@ -715,9 +720,9 @@ class GridFSBucket(object):
         .. versionchanged:: 3.6
            Added ``session`` parameter.
         """
-        gout = self.open_download_stream(file_id, session=session)
-        for chunk in gout:
-            destination.write(chunk)
+        with self.open_download_stream(file_id, session=session) as gout:
+            for chunk in gout:
+                destination.write(chunk)
 
     def delete(self, file_id, session=None):
         """Given an file_id, delete this stored file's files collection document
@@ -890,10 +895,10 @@ class GridFSBucket(object):
         .. versionchanged:: 3.6
            Added ``session`` parameter.
         """
-        gout = self.open_download_stream_by_name(
-            filename, revision, session=session)
-        for chunk in gout:
-            destination.write(chunk)
+        with self.open_download_stream_by_name(
+                filename, revision, session=session) as gout:
+            for chunk in gout:
+                destination.write(chunk)
 
     def rename(self, file_id, new_filename, session=None):
         """Renames the stored file with the specified file_id.
