@@ -15,12 +15,11 @@
 """Test the read_concern module."""
 
 from bson.son import SON
-from pymongo import monitoring
 from pymongo.errors import ConfigurationError, OperationFailure
 from pymongo.read_concern import ReadConcern
 
 from test import client_context, PyMongoTestCase
-from test.utils import single_client, rs_or_single_client, EventListener
+from test.utils import single_client, rs_or_single_client, OvertCommandListener
 
 
 class TestReadConcern(PyMongoTestCase):
@@ -28,17 +27,10 @@ class TestReadConcern(PyMongoTestCase):
     @classmethod
     @client_context.require_connection
     def setUpClass(cls):
-        cls.listener = EventListener()
-        cls.saved_listeners = monitoring._LISTENERS
-        # Don't use any global subscribers.
-        monitoring._LISTENERS = monitoring._Listeners([], [], [], [])
+        cls.listener = OvertCommandListener()
         cls.client = single_client(event_listeners=[cls.listener])
         cls.db = cls.client.pymongo_test
 
-    @classmethod
-    def tearDownClass(cls):
-        monitoring._LISTENERS = cls.saved_listeners
-
     def tearDown(self):
         self.db.coll.drop()
         self.listener.results.clear()
@@ -118,8 +110,14 @@ class TestReadConcern(PyMongoTestCase):
         except OperationFailure:
             # "ns doesn't exist"
             pass
-        self.assertNotIn('readConcern',
-                         self.listener.results['started'][0].command)
+
+        # Aggregate with $out supports readConcern MongoDB 4.2 onwards.
+        if client_context.version >= (4, 1):
+            self.assertIn('readConcern',
+                          self.listener.results['started'][0].command)
+        else:
+            self.assertNotIn('readConcern',
+                          self.listener.results['started'][0].command)
 
     def test_map_reduce_out(self):
         coll = self.db.get_collection('coll', read_concern=ReadConcern('local'))
