@@ -15,6 +15,7 @@
 """Test the client_session module."""
 
 import copy
+import os
 import sys
 
 from bson import DBRef
@@ -28,8 +29,11 @@ from pymongo.errors import (ConfigurationError,
 from pymongo.monotonic import time as _time
 from pymongo.read_concern import ReadConcern
 from test import IntegrationTest, client_context, db_user, db_pwd, unittest, SkipTest
-from test.utils import ignore_deprecations, rs_or_single_client, EventListener
-
+from test.utils import (ignore_deprecations,
+                        rs_or_single_client,
+                        EventListener,
+                        TestCreator)
+from test.utils_spec_runner import SpecRunner
 
 # Ignore auth commands like saslStart, so we can assert lsid is in all commands.
 class SessionTestListener(EventListener):
@@ -816,14 +820,18 @@ class TestCausalConsistency(unittest.TestCase):
                 self.assertEqual(sess.cluster_time, sess2.cluster_time)
                 self.assertEqual(sess.operation_time, sess2.operation_time)
 
-    def _test_reads(self, op):
+    def _test_reads(self, op, exception=None):
         coll = self.client.pymongo_test.test
         with self.client.start_session() as sess:
             coll.find_one({}, session=sess)
             operation_time = sess.operation_time
             self.assertIsNotNone(operation_time)
             self.listener.results.clear()
-            op(coll, sess)
+            if exception:
+                with self.assertRaises(exception):
+                    op(coll, sess)
+            else:
+                op(coll, sess)
             act = self.listener.results['started'][0].command.get(
                 'readConcern', {}).get('afterClusterTime')
             self.assertEqual(operation_time, act)
@@ -844,12 +852,19 @@ class TestCausalConsistency(unittest.TestCase):
             lambda coll, session: coll.count_documents({}, session=session))
         self._test_reads(
             lambda coll, session: coll.distinct('foo', session=session))
+
+        # SERVER-40938 removed support for casually consistent mapReduce.
+        map_reduce_exc = None
+        if client_context.version.at_least(4, 1, 12):
+            map_reduce_exc = OperationFailure
         self._test_reads(
             lambda coll, session: coll.map_reduce(
-                'function() {}', 'function() {}', 'inline', session=session))
+                'function() {}', 'function() {}', 'inline', session=session),
+            exception=map_reduce_exc)
         self._test_reads(
             lambda coll, session: coll.inline_map_reduce(
-                'function() {}', 'function() {}', session=session))
+                'function() {}', 'function() {}', session=session),
+            exception=map_reduce_exc)
         if (not client_context.is_mongos and
                 not client_context.version.at_least(4, 1, 0)):
             def scan(coll, session):
@@ -984,9 +999,6 @@ class TestCausalConsistency(unittest.TestCase):
             lambda coll, session: coll.drop_index("foo_1", session=session))
         self._test_no_read_concern(
             lambda coll, session: coll.drop_indexes(session=session))
-        self._test_no_read_concern(
-                lambda coll, session: list(
-                    coll.aggregate([{"$out": "aggout"}], session=session)))
         self._test_no_read_concern(
             lambda coll, session: coll.map_reduce(
                 'function() {}', 'function() {}', 'mrout', session=session))
@@ -1002,6 +1014,13 @@ class TestCausalConsistency(unittest.TestCase):
             self._test_no_read_concern(
                 lambda coll, session: coll.reindex(session=session))
 
+    @client_context.require_no_standalone
+    @client_context.require_version_max(4, 1, 0)
+    def test_aggregate_out_does_not_include_read_concern(self):
+        self._test_no_read_concern(
+                lambda coll, session: list(
+                    coll.aggregate([{"$out": "aggout"}], session=session)))
+
     @client_context.require_no_standalone
     def test_get_more_does_not_include_read_concern(self):
         coll = self.client.pymongo_test.test
@@ -1038,6 +1057,7 @@ class TestCausalConsistency(unittest.TestCase):
             self.assertIsNone(act)
 
     @client_context.require_no_standalone
+    @client_context.require_no_mmap
     def test_read_concern(self):
         with self.client.start_session(causal_consistency=True) as s:
             coll = self.client.pymongo_test.test
@@ -1291,3 +1311,53 @@ class TestClusterTime(IntegrationTest):
                         succeeded.reply['$clusterTime']['clusterTime'],
                         "%s sent wrong $clusterTime with %s" % (
                             f.__name__, event.command_name))
+
+
+class TestSpec(SpecRunner):
+    # Location of JSON test specifications.
+    TEST_PATH = os.path.join(
+        os.path.dirname(os.path.realpath(__file__)), 'sessions')
+
+    def last_two_command_events(self):
+        """Return the last two command started events."""
+        started_events = self.listener.results['started'][-2:]
+        self.assertEqual(2, len(started_events))
+        return started_events
+
+    def assert_same_lsid_on_last_two_commands(self):
+        """Run the assertSameLsidOnLastTwoCommands test operation."""
+        event1, event2 = self.last_two_command_events()
+        self.assertEqual(event1.command['lsid'], event2.command['lsid'])
+
+    def assert_different_lsid_on_last_two_commands(self):
+        """Run the assertDifferentLsidOnLastTwoCommands test operation."""
+        event1, event2 = self.last_two_command_events()
+        self.assertNotEqual(event1.command['lsid'], event2.command['lsid'])
+
+    def assert_session_dirty(self, session):
+        """Run the assertSessionDirty test operation.
+
+        Assert that the given session is dirty.
+        """
+        self.assertIsNotNone(session._server_session)
+        self.assertTrue(session._server_session.dirty)
+
+    def assert_session_not_dirty(self, session):
+        """Run the assertSessionNotDirty test operation.
+
+        Assert that the given session is not dirty.
+        """
+        self.assertIsNotNone(session._server_session)
+        self.assertFalse(session._server_session.dirty)
+
+
+def create_test(scenario_def, test, name):
+    @client_context.require_test_commands
+    def run_scenario(self):
+        self.run_scenario(scenario_def, test)
+
+    return run_scenario
+
+
+test_creator = TestCreator(create_test, TestSpec, TestSpec.TEST_PATH)
+test_creator.create_tests()
